        <tr class="clickable" onclick="$('#pb_{id}_details').slideToggle(200);">
            <td>{id}</td>
            <td><span id="pb_{id}_n">-</span> / <span id="pb_{id}_total">-</span></td>
            <td>
                <div id="pb_{id}_container" class="pb_container">
                    <div id="pb_{id}" class="pb" style="width: 0"></div>
                </div>
            </td>
            <td><span id="pb_{id}_duration">-</span></td>
            <td><span id="pb_{id}_flash" class="glyphicon glyphicon-flash text-danger"></span>
                <span id="pb_{id}_remaining">-</span></td>
            <td><span id="pb_{id}_started">-</span></td>
            <td><span id="pb_{id}_finished">-</span></td>
        </tr>
        <tr >
            <td colspan="7" class="pb_details">
                <div id="pb_{id}_details">
                    <div class="pb_details_left_filler"></div>
                    <div class="pb_details_right">
                        <h1 class="separator">Task details</h1>
                        <p>Function: <code>{function_name}</code>, on line <code>{function_line_no}</code>
                            of <code>{user}{function_filename}</code></p>
                        <p>Invoked on line <code>{invoked_line_no}</code> of <code>{invoked_filename}</code>,
                            through <code>{invoked_code_context}</code></p>
                        <pre id="pb_{id}_traceback" class="traceback"></pre>
                    </div>

                    <div class="pb_details_left_filler" style="display:{has_insights}"></div>
                    <div class="pb_details_right" style="display:{has_insights}">
                        <h1 class="separator clickable" title="Click to expand/collapse" onclick="
                            $('#pb_{id}_insights').slideToggle(200);
                            $('#pb_{id}_insights_expand').text(($('#pb_{id}_insights_expand').text() == 'expand') ?
                                                               'collapse' : 'expand')
                        ">Insights (click to&nbsp;<span id="pb_{id}_insights_expand">expand</span>)</h1>
                        <div id="pb_{id}_insights" class="insights">
                            <p class="info"><span class="info">&#9432;</span> Start up time denotes the time to spin up
                                a worker. Init time is the time a worker spends on the initialization function, when
                                provided. Waiting time is the time a worker needs to wait for new tasks to come in.
                                Working time is the time a worker spends on the task at hand. Exit time is the time a
                                worker spends on the exit function, when provided.</p>
                            <div class="insights-left">
                                <h1 class="separator">Global stats</h1>
                                <table>
                                    <thead>
                                    <tr>
                                        <th></th>
                                        <th>Total</th><th>Mean</th><th>Std</th><th>Ratio (%)</th>
                                    </tr>
                                    </thead>
                                    <tr>
                                        <th>Start up time</th>
                                        <td id="pb_{id}_insights_total_start_up_time"></td>
                                        <td id="pb_{id}_insights_start_up_time_mean"></td>
                                        <td id="pb_{id}_insights_start_up_time_std"></td>
                                        <td id="pb_{id}_insights_start_up_ratio"></td>
                                    </tr>
                                    <tr>
                                        <th>Init time</th>
                                        <td id="pb_{id}_insights_total_init_time"></td>
                                        <td id="pb_{id}_insights_init_time_mean"></td>
                                        <td id="pb_{id}_insights_init_time_std"></td>
                                        <td id="pb_{id}_insights_init_ratio"></td>
                                    </tr>
                                    <tr>
                                        <th>Waiting time</th>
                                        <td id="pb_{id}_insights_total_waiting_time"></td>
                                        <td id="pb_{id}_insights_waiting_time_mean"></td>
                                        <td id="pb_{id}_insights_waiting_time_std"></td>
                                        <td id="pb_{id}_insights_waiting_ratio"></td>
                                    </tr>
                                    <tr>
                                        <th>Working time</th>
                                        <td id="pb_{id}_insights_total_working_time"></td>
                                        <td id="pb_{id}_insights_working_time_mean"></td>
                                        <td id="pb_{id}_insights_working_time_std"></td>
                                        <td id="pb_{id}_insights_working_ratio"></td>
                                    </tr>
                                    <tr>
                                        <th>Exit time</th>
                                        <td id="pb_{id}_insights_total_exit_time"></td>
                                        <td id="pb_{id}_insights_exit_time_mean"></td>
                                        <td id="pb_{id}_insights_exit_time_std"></td>
                                        <td id="pb_{id}_insights_exit_ratio"></td>
                                    </tr>
                                </table>

                                <h1 class="separator">Task stats</h1>
                                <p class="info"><span class="info">&#9432;</span> This section shows the top 5 tasks
                                    based on duration and is updated every 2 seconds.</p>
                                <table class="insights_table">
                                    <thead>
                                    <tr>
                                        <th></th><th>Time</th><th>Arguments</th>
                                    </tr>
                                    </thead>
                                    <tr id="pb_{id}_insights_task_0" style="display:none">
                                        <td>1.</td>
                                        <td id="pb_{id}_insights_task_0_duration"></td>
                                        <td id="pb_{id}_insights_task_0_args" class="code"></td>
                                    </tr>
                                    <tr id="pb_{id}_insights_task_1" style="display:none">
                                        <td>2.</td>
                                        <td id="pb_{id}_insights_task_1_duration"></td>
                                        <td id="pb_{id}_insights_task_1_args" class="code"></td>
                                    </tr>
                                    <tr id="pb_{id}_insights_task_2" style="display:none">
                                        <td>3.</td>
                                        <td id="pb_{id}_insights_task_2_duration"></td>
                                        <td id="pb_{id}_insights_task_2_args" class="code"></td>
                                    </tr>
                                    <tr id="pb_{id}_insights_task_3" style="display:none">
                                        <td>4.</td>
                                        <td id="pb_{id}_insights_task_3_duration"></td>
                                        <td id="pb_{id}_insights_task_3_args" class="code"></td>
                                    </tr>
                                    <tr id="pb_{id}_insights_task_4" style="display:none">
                                        <td>5.</td>
                                        <td id="pb_{id}_insights_task_4_duration"></td>
                                        <td id="pb_{id}_insights_task_4_args" class="code"></td>
                                    </tr>
                                </table>
                            </div>

                            <div class="insights-middle"></div>

                            <div class="insights-right">
                                <h1 class="separator">Worker stats</h1>
                                <table id="pb_{id}_insights_workers">
                                    <thead>
                                    <tr>
                                        <th>Worker</th><th>Tasks completed</th>
                                        <th title="Total start up time">T. start up time</th>
                                        <th title="Total initialization time">T. init time</th>
                                        <th title="Total waiting time">T. waiting time</th>
                                        <th title="Total working time">T. working time</th>
                                        <th title="Total exit time">T. exit time</th>
                                    </tr>
                                    </thead>
                                    {insights_workers}
                                </table>
                            </div>

                        </div>
                    </div>

                </div>
            </td>
        </tr>